'****************************************************************
'*  Name    : Charlie Antenna Unit.BAS                          *
'*  Author  : John F Drew                                       *
'*  Notice  : Copyright (c) 2005 [John F Drew]                  *
'*          : All Rights Reserved                               *
'*  Date    : 7/1/07                                            *
'*  Version : 1.3                                               *
'*  Notes   : Takes data from ScrewJack and puts AZ             *
'*          : out a serial port if handshaking set              *
'*			: data format A + word + check	   					*
'*			: check is a byte arithmetic sum of 3 data bytes	*
'*			: if SW1 resets count to zero - put jumper to J2    *
'*			: if SW2 add offset of X degrees (see eedata 123Dec)*
'*			: if SW3 reverse count                              *
'*          : this version adds scaling
'****************************************************************

Device 16F628
Config WDT_OFF,PWRTE_ON,HS_OSC,CP_OFF,BODEN_OFF,LVP_OFF,MCLRE_ON 	'set fuses 
Declare XTAL=10
'set up the serial port
HSERIAL_BAUD = 9600        				' Set baud rate to 9600
HSERIAL_RCSTA = %10000000  				' Enable serial port
HSERIAL_TXSTA = %00100000  				' Enable transmit and asynchronous mode 

TRISB=%01001011					  		'(outputs)	RB2, RB4, RB5, RB7
                        		  		'(inputs) 	RB0, RB1, RB3, RB6
TRISA=%11111111             	  		' Configure A (PortA.0-7) as inputs
								  		
Dim Azimuth				As Word			'the azimuth word
Dim Azimuth1            As DWord        'in reversal routine
Dim TempOffset          As DWord        'in offset routine
Dim TotalAZ             As Word         'used to manipulate azimuth outside interrupt
Dim Check				As Word			'checksum out

'following variables used in calculations converting binary to decimal
Dim TempDword           As DWord        'used in bringing to a 4096 reference & scaling
Dim Degrees    			As Word			'used in calculating angle
Dim Step1				As Word			'temp in calculations
Dim PartDeg				As Word			'used in calculating part degrees
Dim Tenths				As Word			'used in calculations part degrees

'following variables used in developing total count for azimuth

Dim TempFloat           As Float        'used in scaling
Dim Divisor             As Word         'for linearity calc
Dim AdditionalError     As DWord        'used in scaling
Dim ConvertDword1       As DWord        'used in scaling
Dim ConvertDword2       As DWord        'used in scaling
Dim TempTotalAZ         As Word         'used in scaling
Dim Difference          As Word         'used in scaling
Dim MemoryPos           As Byte         'used in scaling
Dim TempByte            As Byte         'for linearity calc
Dim TestRemain          As Byte         'for linearity calc
Dim Error0              As Byte         'for linearity calc
Dim Loopcount           As Byte         'loop counter 
Dim CountInMax          As Word         'maximum travel in counts
Dim Error1              As Byte         'for linearity calc
Dim DegSpread           As Word         'degrees covered by this maximum travel count
Dim OffsetB             As Byte         'degrees offset added by switch SW2
Dim Offset              As Word         'count offset eg 90 deg in offsetB = 1024
Dim Direction           As Byte         'direction count
Dim Scaling             As Word         'scaling factor (normally 360 degrees)
Dim ModulusTemp         As Byte         'determines if count led should light

'count in to RB0 the interrupt port, up/down to RA1 and RA0

Symbol Down=PORTA.0                     'pin 17                     'from screwjack
Symbol Up=PORTA.1                       'pin 18                     'from screwjack
Symbol CountIn=PORTB.0                  'pin 6 	 			        'from screwjack 

Symbol ResetZero=PORTA.2  	   			'pin 1                      'SW1 resets count to zero
Symbol SW2=PORTA.3                      'pin 2                      'SW2 if low offsets by deg  set in edata
Symbol SW3=PORTB.1						'pin 7						'SW3 inverts number sent to shack unit

Symbol SerialOut=PORTB.2                'pin 8                      'RS232 out to shack
Symbol MagIncAzLED=PORTB.4				'pin 10                     'lights during data send
Symbol MagDecAzLED=PORTB.5              'pin 11                     'unused
Symbol Handshake=PORTB.6                'pin 12						'handshaking pin, low=send data
Symbol Unused=PORTB.7                   'pin 13                     'connects to J1 pin 8
Symbol SizeEdata = 120                                              '120 eedata bytes used for table

'***************** setup interrupts for pin B0

Symbol INTE = INTCON.4                  ' RB0 Interrupt Enable  1=enable

Symbol INTF = INTCON.1                  ' External Interrupt Flag
Symbol GIE = INTCON.7                   ' Global Interrupt Enable

Symbol INTEDG = OPTION_REG.6            ' Interrupt Edge Select= 1 for rising edge

'***************** load data

EData $00,$00,$00,$01,$01,$01,$01,$02,$02,$02,$03,$03,$03,$03,$04,$04    '120 spots for correction
EData $04,$05,$05,$05,$05,$06,$06,$06,$07,$07,$07,$07,$08,$08,$08,$09    '14 counts per spot
EData $09,$09,$09,$0A,$0A,$0A,$0B,$0B,$0B,$0B,$0C,$0C,$0C,$0D,$0D,$0D    
EData $0D,$0E,$0E,$0E,$0F,$0F,$0F,$0F,$10,$10,$10,$11,$11,$11,$11,$12
EData $12,$12,$13,$13,$13,$13,$14,$14,$14,$15,$15,$15,$15,$16,$16,$16
EData $17,$17,$17,$17,$18,$18,$18,$19,$19,$19,$19,$1A,$1A,$1A,$1B,$1B
EData $1B,$1B,$1C,$1C,$1C,$1C,$1D,$1D,$1D,$1E,$1E,$1E,$1F,$1F,$1F,$1F
EData $20,$20,$20,$21,$21,$21,$21,$22,$22                                'Value 120 need for extrapolation, $00 spare.
ScalingDTA    EData Word 360                                             'stores the scaling over how many degrees
OffsetDTA     EData 90                                                   'Offset triggered by SW2
DegSpreadDTA  EData Word 103                                             'Degrees covered by the chosen CountInMaxDta.
CountInMaxDta EData	Word 1680                                            'Note: Countin MAXIMUM with LSB first in eedata

'CountInMaxDTA must be chosen to give first integer value above true "count maximum" when divided by 120. 
'eg 1628/100 not an integer, but 1680/120 = 14. So choose 1680. 
'Now because we chose 1680 instead of 1628, we must correct DegSpreadDTA 1680*100/1628 =103 to maintain the ratio.

ALL_DIGITAL=true                        'all digital inputs

on_interrupt GetData                    'define the interrupt routine

GoTo INITIALISE                         'jump over sub routines

'**************************************************************************** 
' This is the interrupt routine caused by a +ve edge on Port B.0            *
' which is the count in pin 6 of the PIC                                    *
' The initiate_interrupt routine is called once to establish registers      *
'****************************************************************************	

GetData:  
       If Down = 0 Then                   ' change Direction only if a button pressed
         Direction = 0                    ' Down
       ElseIf Up = 0 Then
         Direction = 1                    ' Up
       EndIf

       If Direction = 0 Then              ' going Down
         If Azimuth > 0 Then              ' cannot go below 0
            Dec Azimuth                   ' update CountIn
         EndIf
       Else                               ' going Up
         If Azimuth < CountInMax Then     ' cannot go above countinMax
            Inc Azimuth                   ' update CountIn
         EndIf 
       EndIf    
 
       INTF=0                             'reset the port RB0 interrupt flag
	   Context Restore                    'go back to normal program       
       
'initiate the interrupt
Initiate_interrupt:
				   GIE=0                  'turn off interrupts while initiating
				   While GIE = 1 : GIE = 0 : Wend        ' ensure its off
				   INTF = 0               'Clear interrupt flag
				   INTE = 1               'enable the RB0 interrupt
				   INTEDG = 1             'interrupt on rising edge
				   GIE=1                  'enable global interrupts
      Return
      
'******************************************************************************* 
' CorrectLinearity corrects for errors by reference to the table.              *
' ConvertErrorSign extracts a signed error -127 to +127 from one byte,         *
' the MSB is used to indicate if a negative or a positive error.               *
' It is likely that all errors will be positive but software allows otherwise  *
'*******************************************************************************
		  
ConvertErrorSign:                          'necessary because errors stored as 127 bit 
      If Error0 > 128 Then                 'numbers plus a sign bit at Bit7
         Error0 = Error0 - 128             'numbers > = 128 are negative
         ConvertDword1 = 0 - Error0        'extracted the error correction and the sign
      Else
         ConvertDword1 = Error0
      EndIf
      
      If Error1 > 128 Then                 'numbers > 0 are negative
         Error1 = Error1 - 128
         ConvertDword2 = 0 - Error1        'extracted the error correction and the sign
      Else
         ConvertDword2 = Error1            'convertdwords now have signed error
      EndIf
            
      Return

CorrectLinearity:                         'corrects for non linearity by checking eedata table
    Divisor = CountInMax / SizeEdata      'scale to size of eedata -divisor is the spread from error0 to error1
    MemoryPos = TotalAZ / Divisor         'determine the step through eedata
    Error0 = ERead MemoryPos              'extract the error byte
    
    TempTotalAZ = MemoryPos * Divisor     'this is the count equating to the exact memory location   
    Difference = TotalAZ - TempTotalAZ    'difference between exact count to memory and the real value of AZ
    
    Inc MemoryPos
    Error1 = ERead MemoryPos              'extract the next one up
    GoSub ConvertErrorSign

    TempDword = ConvertDword2 - ConvertDword1                   'the effective error across error span
    TempFloat = TempDword * Difference / Divisor                'get proportion of error
    AdditionalError =  TempFloat                                'convert back to a rounded integer
    TotalAZ = TotalAZ + ConvertDword1 + AdditionalError         'the total az corrected for error    

    Return

'**************************************************************************** 
' This next routine converts the screwjacks counts to a base that           *
' is usable by one of the definitions in the shack unit i.e. 0-4096         *
' Essentially it simulates an AS5045                                        *
'****************************************************************************	

ConvertToBase4096:                             'convert 1628 over 100 deg to 4096 over scaling deg
    TempDword = TotalAZ * 4096 / Scaling       'needed for compatibility to shack unit 
    TempDword = TempDword * DegSpread          
    TotalAZ = TempDword / CountInMax           'now contains a value from 0-4096 for scaling deg

    Return

'**************************************************************************** 
' This routine converts the raw number to degrees and decimal degrees       *
' for outputting to a serial program such as Hyperterm                      *
' It is not used in Charlie's version as SW2 used elsewhere                 *
'****************************************************************************	


'CONVERT12BITS:                            
''convert 12 bits to decimal degrees       'some arithmetic to convert 0-4096 to degrees
'		  Step1=TotalAZ * 45              'and decimal degrees with integer arithmetic
'		  Degrees=Step1/512               'the whole degrees now in Degrees
'		  PartDeg=Step1 // 512		   
'		  PartDeg=PartDeg * 10
'		  Tenths=PartDeg/512              'now we've got tenths
'		  Step1=PartDeg // 512
'		  If Step1>255 Then               'if its more than 0.05 then increment
'		     If Tenths=9 Then
'		                 Inc Degrees
'		                 Tenths=0
'		     Else
'			 	 		 Inc Tenths
'			 EndIf
'		  EndIf	             
'		  Return
		  		  
'**************************************************************************** 
' initialise variables and setup the interrupt registers                    *
'                                                                           *
'****************************************************************************	


INITIALISE:		  
          DelayMS 1000                                              'wait while things settle
          MagDecAzLED = 0                                           'turn LED2 off
          Loopcount=1                                               'set loop counter as 1 (unchanged by this prog)
          Azimuth=0                                                 'cumulative azimuth count
          OffsetB= ERead OffsetDTA                                  'get offset in degrees
          TempOffset = 512 * OffsetB / 45                           'use DWord as function may exceed 65535
          Offset = TempOffset                                       'the count offset for adding with SW2
          CountInMax = ERead CountInMaxDta                          'get the maximum count
          DegSpread = ERead DegSpreadDTA                            'get the synthetic degrees
          Scaling = ERead ScalingDTA                                'get the scaling factor
          GoSub Initiate_interrupt                                  'get the interrupt for RB0 running
          
'**************************************************************************** 
'                                                                           *
' Main routine                                                              *
'                                                                           *
'****************************************************************************	

MAIN:
		  If ResetZero = 0 Then Azimuth = 0                         'reset azimuth variable with switch.
          
          While GIE = 1 : GIE = 0 : Wend                            'ensure global interrupts off to avoid split word variable
          TotalAZ = Azimuth                                         'get value for azimuth without corruption
          GIE = 1                                                   'turn interrupts back on
          
          GoSub CorrectLinearity                                    'fix the non linear nature of screwjack
          GoSub ConvertToBase4096                                   'base the count on a standard 4096 for 12 bit
		  If SW3=0 Then                                             'invert the numbers
	   	     Azimuth1 = TotalAZ - 4096                              'to allow for reverse rotation
	   	     Azimuth1 = Abs Azimuth1                                'get rid of minus sign     
             TotalAZ=Azimuth1                                       're-assign to word variable used for modified AZ
	   	  EndIf
	   	  
	   	  If SW2=0 Then                                             'add an offset of 90 degrees
             TotalAZ = TotalAZ + Offset                             '1024 is 90 deg in 4096 system
	   	     If TotalAZ > 4096 Then                                 'in case it goes over 4096
                TotalAZ = TotalAZ - 4096                            'bring it back to 0-4096 range
             EndIf   
	   	  EndIf
          
          ModulusTemp = TotalAZ // 2                                'switches LED2 with counting action
          If ModulusTemp = 1 Then 
             MagDecAzLED = 1                                        'LED2 on
          Else MagDecAzLED = 0                                      'LED2 off
          EndIf   
          
          If Handshake=0  Then                                      'shack wants data
             MagIncAzLED = 1                                        'turn on the LED to indicate comms with shack
             Check=65+ TotalAZ + Loopcount                          'calculate checksum
             HRSOut "A",TotalAZ,Loopcount,Check				        'send shackunit data out
		  EndIf
		  
		  While Handshake=0: Wend                                   'wait for handshake to go off
          MagIncAzLED = 0                                           'LED off   
		  
		  GoTo MAIN                                                 'repeat the loop
End
		   	 
		   
